#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
from datetime import date, time, datetime

from .exceptions import NotSupportedError, Error
from .query_result_parser import QueryResultParser
from .query_submitter import QuerySubmitter


class SalesforceCDPCursor:
    """
    This class represents the cursor
    """

    _TRANSLATION_TABLE = str.maketrans({"\\": r"\\",
                                        "\n": r"\\n",
                                        "\r": r"\\r",
                                        "'": r"\'"})

    def __init__(self, connection):
        self.arraysize = 1
        self.description = None
        self.data = None
        self.connection = connection
        self.current_query = None
        self.has_next = None
        self.next_batch_id = None
        self.closed = False
        self.has_result = False

    def execute(self, query, params=None):
        """
        Executes the query and loads the first batch of results. Supports only read only queries
        :param query: The input query.
        :param params: The parameters to the query
        :return: None
        """
        self._check_cursor_closed()
        self.current_query = self._resolve_query_with_params(query, params)
        json_results = QuerySubmitter.execute(self.connection, self.current_query)
        results = QueryResultParser.parse_result(json_results)
        self.description = results.description
        self.data = results.data
        self.has_next = results.has_next
        self.next_batch_id = results.next_batch_id
        self.has_result = True

    def fetchall(self):
        """
        To be called after the execute function. This will return the entire results of the query executed.
        :return: Returns the query results
        """
        if not self.has_result:
            raise Error('No results available to fetch')
        while self.has_next is True:
            self._check_cursor_closed()
            json_results = QuerySubmitter.get_next_batch(self.connection, self.next_batch_id)
            results = QueryResultParser.parse_result(json_results)
            self.description = results.description
            self.has_next = results.has_next
            self.next_batch_id = results.next_batch_id
            self.data = self.data + results.data
        self._check_cursor_closed()
        self.has_result = False
        return self.data

    def fetchone(self):
        """
        To be called after the execute function. This will return next row from the query results.
        :return: Returns the next row from query result. Returns None if no more data is present.
        """
        if not self.has_result:
            raise Error('No results available to fetch')
        self._check_cursor_closed()
        if self.data is not None and len(self.data) > 0:
            next_row = self.data.pop(0)
            return next_row
        elif self.has_next is True:
            json_results = QuerySubmitter.get_next_batch(self.connection, self.next_batch_id)
            results = QueryResultParser.parse_result(json_results)
            self.description = results.description
            self.data = results.data
            self.has_next = results.has_next
            self.next_batch_id = results.next_batch_id
            next_row = self.data.pop(0)
            return next_row
        else:
            self.has_result = False
            return None

    def fetchmany(self, size=None):
        """
        Returns size number of rows from result
        :param size: If not provided cursor.arraysize will be used by default
        :return: size number of rows from result
        """
        if size is None:
            size = self.arraysize
        result = []
        while size > 0:
            next_row = self.fetchone()
            size = size - 1
            if next_row is None:
                break
            result.append(next_row)
        return result

    def close(self):
        """
        Nothing to close
        :return: None
        """
        self.closed = True

    def rollback(self):
        """
        Nothing to rollback
        :return: None
        """
        pass

    def commit(self):
        """
        Nothing to commit
        :return: None
        """
        pass

    def setinputsizes(self):
        """
        Do nothing
        :return:
        """
        pass

    def setoutputsize(self):
        """
        Do nothing
        :return:
        """
        pass

    def _check_cursor_closed(self):
        if self.closed:
            raise Error('Attempting operation while cursor is closed')
        elif self.connection.closed:
            raise Error('Attempting operation while connection is closed')

    def _resolve_query_with_params(self, query, params):
        params_count_from_query = query.count('?')
        if params is None and params_count_from_query == 0:
            return query

        if self._is_iterable(params) and params_count_from_query == len(params):
            for param in params:
                query = self._replace_next_param(query, param)
        elif params_count_from_query == 1 and params is not None:
            query = self._replace_next_param(query, params)
        else:
            raise Exception('Parameter count not matching')

        return query

    def _replace_next_param(self, query, param):
        if param is None:
            query = query.replace('?', 'null', 1)
        elif self._is_numeric(param):
            query = query.replace('?', str(param), 1)
        else:
            if not isinstance(param, str):
                param = str(param)
            param = param.translate(self._TRANSLATION_TABLE)
            query = query.replace('?', f"'{str(param)}'", 1)
        return query

    def _is_iterable(self, param):
        try:
            iter(param)
            return True
        except Exception as e:
            pass
        return False

    def _is_numeric(self, param):
        return isinstance(param, int) or \
            isinstance(param, float)

    def executemany(self, **kwargs):
        raise NotSupportedError('executemany is not supported')


def Date(year, month, day):
    """Construct an object holding a date value."""
    return date(year, month, day)


def Time(hour, minute=0, second=0, microsecond=0, tzinfo=None):
    """Construct an object holding a time value."""
    return time(hour, minute, second, microsecond, tzinfo)


def Timestamp(year, month, day, hour=0, minute=0, second=0, microsecond=0,
              tzinfo=None):
    """Construct an object holding a time stamp value."""
    return datetime(year, month, day, hour, minute, second, microsecond,
                    tzinfo)


def DateFromTicks(ticks):
    return date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    return time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    return datetime(*time.localtime(ticks)[:6])


class Binary(bytes):
    """Construct an object capable of holding a binary (long) string value."""


class DBAPITypeObject:
    def __init__(self, *values):
        self.values = [v.lower() for v in values]

    def __eq__(self, other):
        return other.lower() in self.values


STRING = DBAPITypeObject("VARCHAR", "CHAR", "VARBINARY", "JSON")

NUMBER = DBAPITypeObject(
    "BOOLEAN", "TINYINT", "SMALLINT", "INTEGER", "BIGINT",  "DOUBLE", "DECIMAL"
)

DATETIME = DBAPITypeObject(
    "DATE",
    "TIME",
    "DATETIME",
    "TIMESTAMP"
)




#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
from .authentication_helper import AuthenticationHelper
from .constants import API_VERSION_V2
from .cursor import SalesforceCDPCursor
from .exceptions import Error
from .pandas_utils import PandasUtils

apilevel = "2.0"
threadsafety = 2
paramstyle = "qmark"


class SalesforceCDPConnection:
    """
    This object represents a connection to CDP
    """

    def __init__(self, login_url, username=None, password=None, client_id=None, client_secret=None,
                 api=API_VERSION_V2, core_token=None, refresh_token=None):
        self.login_url = login_url
        self.username = username
        self.password = password
        self.client_id = client_id
        self.client_secret = client_secret
        self.api = api
        self.core_token = core_token
        self.refresh_token = refresh_token
        self.closed = False
        self.authentication_helper = AuthenticationHelper(self)

    def cursor(self):
        """
        :return: Returns a new instance of the SalesforceCDPCursor
        """
        if self.closed:
            raise Error('Cannot create cursor. Connection is closed')
        return SalesforceCDPCursor(self)

    def get_pandas_dataframe(self, query):
        """
        Returns the query result as Pandas Dataframe
        :param query: The input query
        :return: Query Results as Pandas dataframe
        """
        if self.closed:
            raise Error('Cannot create dataframe. Connection is closed')
        return PandasUtils.get_dataframe(self, query)

    def close(self):
        """
        Clearing credentials
        Marking connection as closed
        :return: None
        """
        self.login_url = None
        self.username = None
        self.password = None
        self.client_id = None
        self.client_secret = None
        self.core_token = None
        self.refresh_token = None
        self.authentication_helper = None
        self.closed = True

    def commit(self):
        """
        Nothing to commit
        :return: None
        """
        pass

    def rollback(self):
        """
        Nothing to rollback
        :return: None
        """
        pass



#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json

import requests

from .constants import *
from .exceptions import Error


class QuerySubmitter:
    """
    Helper methods to execute query against V2 API
    """

    @staticmethod
    def execute(connection, query, api_version=API_VERSION_V2, enable_arrow_stream=False):
        """
        This method submits the query to queryV2 API for execution
        :param connection: SalesforceCDPConnection
        :param query: The query to be executed
        :param api_version: v1 or v2 API
        :param enable_arrow_stream: Set as True to fetch the results as ArrowStream
        :return: Returns the response JSON
        """
        token, instance_url = connection.authentication_helper.get_token()
        return QuerySubmitter._get_query_results(query, instance_url, token, api_version, enable_arrow_stream)

    @staticmethod
    def get_next_batch(connection, next_batch_id, enable_arrow_stream=False):
        """
        This method fetches the next batch of results using the v2 APIs.
        :param connection:  SalesforceCDPConnection
        :param next_batch_id: batchId to fetch the results
        :param enable_arrow_stream: Set as True to fetch the results as ArrowStream
        :return:
        """
        token, instance_url = connection.authentication_helper.get_token()
        return QuerySubmitter._get_next_batch_results(next_batch_id, instance_url, token, enable_arrow_stream)

    @staticmethod
    def _get_query_results(query, instance_url, token, api_version='V2', enable_arrow_stream=False):
        url = f'https://{instance_url}/api/{api_version}/query'
        json_payload = QuerySubmitter._get_payload(query)
        headers = QuerySubmitter._get_headers(token, enable_arrow_stream)
        sql_response = requests.post(url=url, data=json_payload, headers=headers, verify=False)
        if sql_response.status_code != 200:
            try:
                error_json = sql_response.json()
                error_message = error_json['message']
            finally:
                if error_message is not None:
                    raise Error('Failed executing query in server : %s' % error_message)
                raise Error('Failed executing query in server')
        response_json = sql_response.json()
        return response_json

    @staticmethod
    def _get_next_batch_results(next_batch_id, instance_url, token, enable_arrow_stream=False):
        url = f'https://{instance_url}/api/v2/query/{next_batch_id}'
        headers = QuerySubmitter._get_headers(token, enable_arrow_stream)
        sql_response = requests.get(url=url, headers=headers, verify=False)
        response_json = sql_response.json()
        if sql_response.status_code != 200:
            try:
                error_json = sql_response.json()
                error_message = error_json['message']
            finally:
                if error_message is not None:
                    raise Error('Failed executing query in server : %s' % error_message)
                raise Error('Failed executing query in server')
        return response_json

    @staticmethod
    def _get_headers(token, enable_arrow_stream):
        headers = {QUERY_HEADER_KEY_AUTHORIZATION: f'Bearer {token}',
                   QUERY_HEADER_KEY_CONTENT_TYPE: QUERY_HEADER_VALUE_APPLICATION_JSON,
                   QUERY_HEADER_KEY_ACCEPT_ENCODING: QUERY_HEADER_VALUE_GZIP}
        if enable_arrow_stream:
            headers['enable-arrow-stream'] = 'true'
        return headers

    @staticmethod
    def _get_payload(query):
        payload = {
            'sql': query
        }
        json_payload = json.dumps(payload)
        return json_payload



#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
import dateutil.parser

from .constants import *
from .parsed_query_result import QueryResult


class QueryResultParser:

    @staticmethod
    def parse_result(result):
        """
        Parses the json response from queryV2 API
        :param result: JSON response from queryV2 API
        :return: ParsedQueryResult
        """
        return QueryResultParser._parse_v2_result(result)

    @staticmethod
    def _parse_v2_result(result):
        data = result[QUERY_RESPONSE_KEY_DATA]
        metadata_dict = result[QUERY_RESPONSE_KEY_METADATA]
        is_done = result[QUERY_RESPONSE_KEY_DONE]
        next_batch_id = result.get(QUERY_RESPONSE_KEY_NEXT_BATCH_ID)
        has_next = not is_done
        sorted_metadata_items = QueryResultParser._sort_metadata_by_place_in_order(metadata_dict)
        description = QueryResultParser._convert_metadata_list_to_description(sorted_metadata_items)
        QueryResultParser._convert_timestamps(data, description)
        return QueryResult(data, description, has_next, next_batch_id)

    @staticmethod
    def _convert_timestamps(data, description):
        """
        TIMESTAMPS are coming as string in JSON. This function will update the string to datetime object
        :param data: List of JSON results
        :param description: Cursor description
        :return: None
        """
        for i in range(0, len(description)):
            if description[i][1] == DATA_TYPE_TIMESTAMP:
                for data_row in data:
                    if data_row[i] is not None and isinstance(data_row[i], str) and len(data_row[i]) > 0:
                        data_row[i] = dateutil.parser.parse(data_row[i])

    @staticmethod
    def _convert_metadata_item_to_description_item(metadata_item):
        metadata_name = metadata_item[0]
        type = metadata_item[1]['type']
        return QueryResultParser._get_description_item(metadata_name, type)

    @staticmethod
    def _get_description_item(name, data_type):
        """
        This will generate the description to be used with Cursor
        :param name: Column Name
        :param data_type: Column Type
        :return: One tuple representing description for a column
        """
        return (
            name, # Column Name
            data_type, # Column Type
            None,
            None,
            None,
            None,
            None
        )

    @staticmethod
    def _convert_metadata_list_to_description(sorted_metadata_list):
        return [QueryResultParser._convert_metadata_item_to_description_item(metadata_item) for metadata_item in
                sorted_metadata_list]

    @staticmethod
    def _sort_metadata_by_place_in_order(metadata_dict):
        """
        This functions sorts the column metadata from queryV2 response based on the placeInOrder field.
        :param metadata_dict: The metadata dict from JSON
        :return: sorted column metadata as List
        """
        metadataList = [metadataItem for metadataItem in metadata_dict.items()]
        metadataList = sorted(metadataList, key=lambda item: item[1][QUERY_RESPONSE_KEY_PLACE_IN_ORDER])
        return metadataList



#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

class QueryResult:

    def __init__(self, data, description, has_next, next_batch_id):
        self.data = data
        self.description = description
        self.has_next = has_next
        self.next_batch_id = next_batch_id



#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
from .constants import *
from .exceptions import Error
from datetime import datetime, timedelta

import requests


class AuthenticationHelper:

    def __init__(self, connection):
        self.exchange_token = None
        self.instance_url = None
        self.token_expiry_time = None
        self.connection = connection

    def get_token(self):
        """
        Retrieves the cdp token and instance url.
        The connection should be having client_id, client_secret and either refresh token or username and password

        :return: cdp token, instance_url
        """
        if self._is_token_valid():
            return self.exchange_token, self.instance_url
        if self.connection.refresh_token is not None:
            if self.connection.core_token is not None:
                try:
                    return self._exchange_token(self.connection.login_url, self.connection.core_token)
                except:
                    # core token might be expired
                    # try renewing token
                    pass

            return self._renew_token(self.connection.login_url, self.connection.refresh_token,
                                     self.connection.client_id, self.connection.client_secret)
        elif self.connection.password is not None:
            return self._token_by_un_pwd_flow(self.connection.login_url, self.connection.client_id,
                                              self.connection.client_secret, self.connection.username,
                                              self.connection.password)
        else:
            raise Error('Sufficient information is not available for authentication')

    def _is_token_valid(self):
        """
        Checks if the token is valid

        :return: True if the token is valid
        """
        if self.token_expiry_time is not None and self.exchange_token is not None:
            current_time = datetime.now()
            return current_time < self.token_expiry_time
        return False

    def _exchange_token(self, login_url, core_token):
        params = {AUTH_PARAM_GRANT_TYPE: AUTH_PARAM_CDP_GRANT_TYPE,
                  AUTH_PARAM_CDP_SUBJECT_TOKEN_TYPE: AUTH_PARAM_CDP_SUBJECT_TOKEN_TYPE_VALUE,
                  AUTH_PARAM_CDP_SUBJECT_TOKEN: core_token}
        current_time = datetime.now()
        access_code_res = requests.post(url=login_url + '/services/a360/token', params=params)
        if access_code_res.status_code == 200:
            access_code = access_code_res.json()
            access_token = access_code[AUTH_RESPONSE_ACCESS_TOKEN]
            expires_in_seconds = access_code[AUTH_RESPONSE_EXPIRES_IN]
            instance_url = access_code[AUTH_RESPONSE_INSTANCE_URL]
            token_expiry_time = current_time + timedelta(seconds=expires_in_seconds)
            AuthenticationHelper._revoke_core_token(login_url, core_token)
        else:
            raise Error('CDP token retrieval failed with code %d' % access_code_res.status_code)
        self.exchange_token = access_token
        self.token_expiry_time = token_expiry_time
        self.instance_url = instance_url
        return access_token, instance_url

    @staticmethod
    def _revoke_core_token(login_url, core_token):
        """
        Revokes the core token
        :param login_url: The login URL
        :param core_token: The core token
        :return:
        """
        params = {'token': core_token}
        access_code_res = requests.post(url=login_url + '/services/oauth2/revoke', params=params)
        if access_code_res.status_code != 200:
            raise Error('Core token revoke failed with code %d' % access_code_res.status_code)

    def _renew_token(self, login_url, refresh_token, client_id, client_secret):
        """
        Revives the expired token

        :param login_url: The login URL
        :param refresh_token: The OAuth refresh token
        :param client_id: The client id for the app
        :param client_secret: The client secret for the app
        :return: cdp_token, instance_url will be returned
        """
        params = {AUTH_PARAM_GRANT_TYPE: AUTH_PARAM_REFRESH_TOKEN_GRANT_TYPE,
                  AUTH_PARAM_CLIENT_ID: client_id,
                  AUTH_PARAM_CLIENT_SECRET: client_secret,
                  AUTH_PARAM_REFRESH_TOKEN_GRANT_TYPE: refresh_token}
        access_code_res = requests.post(url=login_url + '/services/oauth2/token', params=params)
        if access_code_res.status_code == 200:
            access_code = access_code_res.json()
            core_token = access_code[AUTH_RESPONSE_ACCESS_TOKEN]
            return self._exchange_token(login_url, core_token)
        else:
            raise Error('Token Renewal failed with code %d' % access_code_res.status_code)

    def _token_by_un_pwd_flow(self, login_url, client_id, client_secret, username, password):
        """
        This function fetches the core token
        :param login_url: The Login URL for the tenant
        :param client_id: The client id for the app
        :param client_secret: The client secret for the app
        :param username: Tenant username
        :param password: Tenant password
        :return: cdp_token, instance_url will be returned
        """
        params = {AUTH_PARAM_GRANT_TYPE: AUTH_PARAM_P_D, AUTH_PARAM_CLIENT_ID: client_id,
                  AUTH_PARAM_CLIENT_SECRET: client_secret,
                  AUTH_PARAM_USERNAME: username, AUTH_PARAM_P_D: password}
        access_code_res = requests.post(url=login_url + '/services/oauth2/token', params=params)
        if access_code_res.status_code == 200:
            access_code = access_code_res.json()
            core_token = access_code[AUTH_RESPONSE_ACCESS_TOKEN]
            return self._exchange_token(login_url, core_token)
        else:
            raise Error('Core token retrieval failed with code %d' % access_code_res.status_code)



#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

API_VERSION_V1 = "v1"
API_VERSION_V2 = "v2"

AUTH_PARAM_GRANT_TYPE = 'grant_type'
AUTH_PARAM_P_D = 'password'
AUTH_PARAM_CLIENT_ID = 'client_id'
AUTH_PARAM_CLIENT_SECRET = 'client_secret'
AUTH_PARAM_USERNAME = 'username'
AUTH_RESPONSE_ACCESS_TOKEN = 'access_token'
AUTH_RESPONSE_EXPIRES_IN = 'expires_in'
AUTH_RESPONSE_INSTANCE_URL = 'instance_url'

AUTH_PARAM_CDP_GRANT_TYPE = 'urn:salesforce:grant-type:external:cdp'
AUTH_PARAM_CDP_SUBJECT_TOKEN = 'subject_token'
AUTH_PARAM_CDP_SUBJECT_TOKEN_TYPE = 'subject_token_type'
AUTH_PARAM_CDP_SUBJECT_TOKEN_TYPE_VALUE = 'urn:ietf:params:oauth:token-type:access_token'
AUTH_PARAM_CDP_SUBJECT_TOKEN_TYPE_VALUE = 'urn:ietf:params:oauth:token-type:access_token'
AUTH_PARAM_REFRESH_TOKEN_GRANT_TYPE = 'refresh_token'

DATA_TYPE_TIMESTAMP = 'TIMESTAMP'

ENCODING_ASCII = 'ascii'

QUERY_RESPONSE_KEY_ARROW_STREAM = 'arrowStream'
QUERY_RESPONSE_KEY_DATA = 'data'
QUERY_RESPONSE_KEY_METADATA = 'metadata'
QUERY_RESPONSE_KEY_METADATA_TYPE = 'type'
QUERY_RESPONSE_KEY_DONE = 'done'
QUERY_RESPONSE_KEY_NEXT_BATCH_ID = 'nextBatchId'
QUERY_RESPONSE_KEY_PLACE_IN_ORDER = 'placeInOrder'

QUERY_HEADER_KEY_AUTHORIZATION = 'Authorization'
QUERY_HEADER_KEY_CONTENT_TYPE = 'Content-Type'
QUERY_HEADER_KEY_ACCEPT_ENCODING = 'Accept-Encoding'
QUERY_HEADER_VALUE_APPLICATION_JSON = 'application/json'
QUERY_HEADER_VALUE_GZIP = 'gzip'


#
#  Copyright (c) 2022, salesforce.com, inc.
#  All rights reserved.
#  SPDX-License-Identifier: BSD-3-Clause
#  For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import base64

import dateutil.parser
import pyarrow

from .constants import *
from .query_submitter import QuerySubmitter


class PandasUtils:

    @staticmethod
    def get_dataframe(connection, query):
        """
        Executes the query and returns results as a Pandas dataframe
        :param connection: SalesforceCDPConnection object
        :param query: The query to be executed
        :return: Query results as Pandas Dataframe
        """
        arrow_stream_list = []
        result = QuerySubmitter.execute(connection, query, API_VERSION_V2, True)
        encoded_arrow_stream = result[QUERY_RESPONSE_KEY_ARROW_STREAM]
        arrow_table = PandasUtils._get_pyarrow_table(encoded_arrow_stream)
        PandasUtils._add_table_to_list(arrow_stream_list, arrow_table)
        while result[QUERY_RESPONSE_KEY_DONE] is not True:
            result = QuerySubmitter.get_next_batch(connection, result[QUERY_RESPONSE_KEY_NEXT_BATCH_ID], True)
            encoded_arrow_stream = result[QUERY_RESPONSE_KEY_ARROW_STREAM]
            arrow_table = PandasUtils._get_pyarrow_table(encoded_arrow_stream)
            PandasUtils._add_table_to_list(arrow_stream_list, arrow_table)

        if len(arrow_stream_list) > 0:
            pandas_df = pyarrow.concat_tables(arrow_stream_list).to_pandas()
            date_columns = PandasUtils._get_date_columns(result)
            for date_column in date_columns:
                pandas_df = pandas_df.apply(lambda row: PandasUtils._convert_to_date(row, date_column), axis=1)
            return pandas_df

        return None

    @staticmethod
    def _convert_to_date(row, column):
        value = row[column]
        if isinstance(value, str):
            row[column] = dateutil.parser.parse(value)
        else:
            row[column] = None
        return row

    @staticmethod
    def _get_date_columns(result):
        metadata = result[QUERY_RESPONSE_KEY_METADATA]
        date_columns = [x for x in metadata.keys() if PandasUtils._istimestamp(x, metadata)]
        return date_columns

    @staticmethod
    def _istimestamp(key, metadata_list):
        metadata_type = metadata_list[key][QUERY_RESPONSE_KEY_METADATA_TYPE]
        if metadata_type is not None:
            metadata_type = metadata_type.upper()
        return metadata_type == DATA_TYPE_TIMESTAMP

    @staticmethod
    def _get_pyarrow_table(encoded_arrow_stream):
        if encoded_arrow_stream is None:
            return None
        stream_bytes = encoded_arrow_stream.encode(ENCODING_ASCII)
        decoded_bytes = base64.b64decode(stream_bytes)
        return pyarrow.ipc.open_stream(decoded_bytes).read_all()

    @staticmethod
    def _add_table_to_list(arrow_stream_list, arrow_stream):
        if arrow_stream is not None:
            arrow_stream_list.append(arrow_stream)